@@ -3,17 +3,20 @@ import time
 import sys
 import platform
 import queue
-from collections import namedtuple
 from functools import partial
+from typing import NamedTuple, Callable, Optional, TYPE_CHECKING
 
 from PyQt5.QtGui import *
 from PyQt5.QtCore import *
 from PyQt5.QtWidgets import *
 
-from electrum.i18n import _
+from electrum.i18n import _, languages
 from electrum.util import FileImportFailed, FileExportFailed
 from electrum.paymentrequest import PR_UNPAID, PR_PAID, PR_EXPIRED
 
+if TYPE_CHECKING:
+    from .main_window import ElectrumWindow
+
 
 if platform.system() == 'Windows':
     MONOSPACE_FONT = 'Lucida Console'
@@ -45,19 +48,6 @@ expiration_values = [
 ]
 
 
-class Timer(QThread):
-    stopped = False
-    timer_signal = pyqtSignal()
-
-    def run(self):
-        while not self.stopped:
-            self.timer_signal.emit()
-            time.sleep(0.5)
-
-    def stop(self):
-        self.stopped = True
-        self.wait()
-
 class EnterButton(QPushButton):
     def __init__(self, text, func):
         QPushButton.__init__(self, text)
@@ -200,24 +190,24 @@ class MessageBoxMixin(object):
                             parent, title or '',
                             msg, buttons=Yes|No, defaultButton=No) == Yes
 
-    def show_warning(self, msg, parent=None, title=None, **kwargs):
+    def show_warning(self, msg, parent=None, title=None):
         return self.msg_box(QMessageBox.Warning, parent,
-                            title or _('Warning'), msg, **kwargs)
+                            title or _('Warning'), msg)
 
-    def show_error(self, msg, parent=None, **kwargs):
+    def show_error(self, msg, parent=None):
         return self.msg_box(QMessageBox.Warning, parent,
-                            _('Error'), msg, **kwargs)
+                            _('Error'), msg)
 
-    def show_critical(self, msg, parent=None, title=None, **kwargs):
+    def show_critical(self, msg, parent=None, title=None):
         return self.msg_box(QMessageBox.Critical, parent,
-                            title or _('Critical Error'), msg, **kwargs)
+                            title or _('Critical Error'), msg)
 
-    def show_message(self, msg, parent=None, title=None, **kwargs):
+    def show_message(self, msg, parent=None, title=None):
         return self.msg_box(QMessageBox.Information, parent,
-                            title or _('Information'), msg, **kwargs)
+                            title or _('Information'), msg)
 
     def msg_box(self, icon, parent, title, text, buttons=QMessageBox.Ok,
-                defaultButton=QMessageBox.NoButton, rich_text=False):
+                defaultButton=QMessageBox.NoButton):
         parent = parent or self.top_level_window()
         if type(icon) is QPixmap:
             d = QMessageBox(QMessageBox.Information, title, str(text), buttons, parent)
@@ -226,12 +216,7 @@ class MessageBoxMixin(object):
             d = QMessageBox(icon, title, str(text), buttons, parent)
         d.setWindowModality(Qt.WindowModal)
         d.setDefaultButton(defaultButton)
-        if rich_text:
-            d.setTextInteractionFlags(Qt.TextSelectableByMouse| Qt.LinksAccessibleByMouse)
-            d.setTextFormat(Qt.RichText)
-        else:
-            d.setTextInteractionFlags(Qt.TextSelectableByMouse)
-            d.setTextFormat(Qt.PlainText)
+        d.setTextInteractionFlags(Qt.TextSelectableByMouse)
         return d.exec_()
 
 class WindowModalDialog(QDialog, MessageBoxMixin):
@@ -400,156 +385,158 @@ def filename_field(parent, config, defaultname, select_msg):
     return vbox, filename_e, b1
 
 class ElectrumItemDelegate(QStyledItemDelegate):
-    def createEditor(self, parent, option, index):
-        return self.parent().createEditor(parent, option, index)
+    def __init__(self, tv):
+        super().__init__(tv)
+        self.tv = tv
+        self.opened = None
+        def on_closeEditor(editor: QLineEdit, hint):
+            self.opened = None
+        def on_commitData(editor: QLineEdit):
+            new_text = editor.text()
+            idx = QModelIndex(self.opened)
+            row, col = idx.row(), idx.column()
+            _prior_text, user_role = self.tv.text_txid_from_coordinate(row, col)
+            # check that we didn't forget to set UserRole on an editable field
+            assert user_role is not None, (row, col)
+            self.tv.on_edited(idx, user_role, new_text)
+        self.closeEditor.connect(on_closeEditor)
+        self.commitData.connect(on_commitData)
 
-class MyTreeWidget(QTreeWidget):
+    def createEditor(self, parent, option, idx):
+        self.opened = QPersistentModelIndex(idx)
+        return super().createEditor(parent, option, idx)
 
-    def __init__(self, parent, create_menu, headers, stretch_column=None,
-                 editable_columns=None):
-        QTreeWidget.__init__(self, parent)
+class MyTreeView(QTreeView):
+
+    def __init__(self, parent: 'ElectrumWindow', create_menu, stretch_column=None, editable_columns=None):
+        super().__init__(parent)
         self.parent = parent
         self.config = self.parent.config
         self.stretch_column = stretch_column
         self.setContextMenuPolicy(Qt.CustomContextMenu)
         self.customContextMenuRequested.connect(create_menu)
         self.setUniformRowHeights(True)
-        # extend the syntax for consistency
-        self.addChild = self.addTopLevelItem
-        self.insertChild = self.insertTopLevelItem
 
         self.icon_cache = IconCache()
 
         # Control which columns are editable
-        self.editor = None
-        self.pending_update = False
         if editable_columns is None:
             editable_columns = {stretch_column}
         else:
             editable_columns = set(editable_columns)
         self.editable_columns = editable_columns
         self.setItemDelegate(ElectrumItemDelegate(self))
-        self.itemDoubleClicked.connect(self.on_doubleclick)
-        self.update_headers(headers)
         self.current_filter = ""
 
         self.setRootIsDecorated(False)  # remove left margin
         self.toolbar_shown = False
 
+        # When figuring out the size of columns, Qt by default looks at
+        # the first 1000 rows (at least if resize mode is QHeaderView.ResizeToContents).
+        # This would be REALLY SLOW, and it's not perfect anyway.
+        # So to speed the UI up considerably, set it to
+        # only look at as many rows as currently visible.
+        self.header().setResizeContentsPrecision(0)
+
+    def set_editability(self, items):
+        for idx, i in enumerate(items):
+            i.setEditable(idx in self.editable_columns)
+
+    def selected_in_column(self, column: int):
+        items = self.selectionModel().selectedIndexes()
+        return list(x for x in items if x.column() == column)
+
+    def current_item_user_role(self, col) -> Optional[QStandardItem]:
+        idx = self.selectionModel().currentIndex()
+        idx = idx.sibling(idx.row(), col)
+        item = self.model().itemFromIndex(idx)
+        if item:
+            return item.data(Qt.UserRole)
+
+    def set_current_idx(self, set_current: QPersistentModelIndex):
+        if set_current:
+            assert isinstance(set_current, QPersistentModelIndex)
+            assert set_current.isValid()
+            self.selectionModel().select(QModelIndex(set_current), QItemSelectionModel.SelectCurrent)
+
     def update_headers(self, headers):
-        self.setColumnCount(len(headers))
-        self.setHeaderLabels(headers)
+        model = self.model()
+        model.setHorizontalHeaderLabels(headers)
         self.header().setStretchLastSection(False)
         for col in range(len(headers)):
             sm = QHeaderView.Stretch if col == self.stretch_column else QHeaderView.ResizeToContents
             self.header().setSectionResizeMode(col, sm)
 
-    def editItem(self, item, column):
-        if column in self.editable_columns:
-            try:
-                self.editing_itemcol = (item, column, item.text(column))
-                # Calling setFlags causes on_changed events for some reason
-                item.setFlags(item.flags() | Qt.ItemIsEditable)
-                QTreeWidget.editItem(self, item, column)
-                item.setFlags(item.flags() & ~Qt.ItemIsEditable)
-            except RuntimeError:
-                # (item) wrapped C/C++ object has been deleted
-                pass
-
     def keyPressEvent(self, event):
-        if event.key() in [ Qt.Key_F2, Qt.Key_Return ] and self.editor is None:
-            self.on_activated(self.currentItem(), self.currentColumn())
-        else:
-            QTreeWidget.keyPressEvent(self, event)
+        if self.itemDelegate().opened:
+            return
+        if event.key() in [ Qt.Key_F2, Qt.Key_Return ]:
+            self.on_activated(self.selectionModel().currentIndex())
+            return
+        super().keyPressEvent(event)
 
-    def permit_edit(self, item, column):
-        return (column in self.editable_columns
-                and self.on_permit_edit(item, column))
-
-    def on_permit_edit(self, item, column):
-        return True
-
-    def on_doubleclick(self, item, column):
-        if self.permit_edit(item, column):
-            self.editItem(item, column)
-
-    def on_activated(self, item, column):
+    def on_activated(self, idx):
         # on 'enter' we show the menu
-        pt = self.visualItemRect(item).bottomLeft()
+        pt = self.visualRect(idx).bottomLeft()
         pt.setX(50)
         self.customContextMenuRequested.emit(pt)
 
-    def createEditor(self, parent, option, index):
-        self.editor = QStyledItemDelegate.createEditor(self.itemDelegate(),
-                                                       parent, option, index)
-        self.editor.editingFinished.connect(self.editing_finished)
-        return self.editor
+    def edit(self, idx, trigger=QAbstractItemView.AllEditTriggers, event=None):
+        """
+        this is to prevent:
+           edit: editing failed
+        from inside qt
+        """
+        return super().edit(idx, trigger, event)
 
-    def editing_finished(self):
-        # Long-time QT bug - pressing Enter to finish editing signals
-        # editingFinished twice.  If the item changed the sequence is
-        # Enter key:  editingFinished, on_change, editingFinished
-        # Mouse: on_change, editingFinished
-        # This mess is the cleanest way to ensure we make the
-        # on_edited callback with the updated item
-        if self.editor:
-            (item, column, prior_text) = self.editing_itemcol
-            if self.editor.text() == prior_text:
-                self.editor = None  # Unchanged - ignore any 2nd call
-            elif item.text(column) == prior_text:
-                pass # Buggy first call on Enter key, item not yet updated
-            else:
-                # What we want - the updated item
-                self.on_edited(*self.editing_itemcol)
-                self.editor = None
-
-            # Now do any pending updates
-            if self.editor is None and self.pending_update:
-                self.pending_update = False
-                self.on_update()
-
-    def on_edited(self, item, column, prior):
-        '''Called only when the text actually changes'''
-        key = item.data(0, Qt.UserRole)
-        text = item.text(column)
-        self.parent.wallet.set_label(key, text)
-        self.parent.history_list.update_labels()
+    def on_edited(self, idx: QModelIndex, user_role, text):
+        self.parent.wallet.set_label(user_role, text)
+        self.parent.history_model.refresh('on_edited in MyTreeView')
         self.parent.update_completions()
 
-    def update(self):
-        # Defer updates if editing
-        if self.editor:
-            self.pending_update = True
+    def should_hide(self, row):
+        """
+        row_num is for self.model(). So if there is a proxy, it is the row number
+        in that!
+        """
+        return False
+
+    def text_txid_from_coordinate(self, row_num, column):
+        assert not isinstance(self.model(), QSortFilterProxyModel)
+        idx = self.model().index(row_num, column)
+        item = self.model().itemFromIndex(idx)
+        user_role = item.data(Qt.UserRole)
+        return item.text(), user_role
+
+    def hide_row(self, row_num):
+        """
+        row_num is for self.model(). So if there is a proxy, it is the row number
+        in that!
+        """
+        should_hide = self.should_hide(row_num)
+        if not self.current_filter and should_hide is None:
+            # no filters at all, neither date nor search
+            self.setRowHidden(row_num, QModelIndex(), False)
+            return
+        for column in self.filter_columns:
+            txt, _ = self.text_txid_from_coordinate(row_num, column)
+            txt = txt.lower()
+            if self.current_filter in txt:
+                # the filter matched, but the date filter might apply
+                self.setRowHidden(row_num, QModelIndex(), bool(should_hide))
+                break
         else:
-            self.setUpdatesEnabled(False)
-            scroll_pos = self.verticalScrollBar().value()
-            self.on_update()
-            self.setUpdatesEnabled(True)
-            # To paint the list before resetting the scroll position
-            self.parent.app.processEvents()
-            self.verticalScrollBar().setValue(scroll_pos)
-        if self.current_filter:
-            self.filter(self.current_filter)
-
-    def on_update(self):
-        pass
-
-    def get_leaves(self, root):
-        child_count = root.childCount()
-        if child_count == 0:
-            yield root
-        for i in range(child_count):
-            item = root.child(i)
-            for x in self.get_leaves(item):
-                yield x
+            # we did not find the filter in any columns, hide the item
+            self.setRowHidden(row_num, QModelIndex(), True)
 
     def filter(self, p):
-        columns = self.__class__.filter_columns
         p = p.lower()
         self.current_filter = p
-        for item in self.get_leaves(self.invisibleRootItem()):
-            item.setHidden(all([item.text(column).lower().find(p) == -1
-                                for column in columns]))
+        self.hide_rows()
+
+    def hide_rows(self):
+        for row in range(self.model().rowCount()):
+            self.hide_row(row)
 
     def create_toolbar(self, config=None):
         hbox = QHBoxLayout()
@@ -601,6 +588,7 @@ class ButtonsWidget(QWidget):
     def addButton(self, icon_name, on_click, tooltip):
         button = QToolButton(self)
         button.setIcon(QIcon(icon_name))
+        button.setIconSize(QSize(25,25))
         button.setStyleSheet("QToolButton { border: none; hover {border: 1px} pressed {border: 1px} padding: 0px; }")
         button.setVisible(True)
         button.setToolTip(tooltip)
@@ -643,7 +631,12 @@ class TaskThread(QThread):
     '''Thread that runs background tasks.  Callbacks are guaranteed
     to happen in the context of its parent.'''
 
-    Task = namedtuple("Task", "task cb_success cb_done cb_error")
+    class Task(NamedTuple):
+        task: Callable
+        cb_success: Optional[Callable]
+        cb_done: Optional[Callable]
+        cb_error: Optional[Callable]
+
     doneSig = pyqtSignal(object, object, object)
 
     def __init__(self, parent, on_error=None):
@@ -659,7 +652,7 @@ class TaskThread(QThread):
 
     def run(self):
         while True:
-            task = self.tasks.get()
+            task = self.tasks.get()  # type: TaskThread.Task
             if not task:
                 break
             try:
@@ -790,22 +783,6 @@ def get_parent_main_window(widget):
             return widget
     return None
 
-class SortableTreeWidgetItem(QTreeWidgetItem):
-    DataRole = Qt.UserRole + 1
-
-    def __lt__(self, other):
-        column = self.treeWidget().sortColumn()
-        if None not in [x.data(column, self.DataRole) for x in [self, other]]:
-            # We have set custom data to sort by
-            return self.data(column, self.DataRole) < other.data(column, self.DataRole)
-        try:
-            # Is the value something numeric?
-            return float(self.text(column)) < float(other.text(column))
-        except ValueError:
-            # If not, we will just do string comparison
-            return self.text(column) < other.text(column)
-
-
 class IconCache:
 
     def __init__(self):
@@ -817,6 +794,26 @@ class IconCache:
         return self.__cache[file_name]
 
 
+def get_default_language():
+    name = QLocale.system().name()
+    return name if name in languages else 'en_UK'
+
+class FromList(QTreeWidget):
+    def __init__(self, parent, create_menu):
+        super().__init__(parent)
+        self.setHeaderHidden(True)
+        self.setMaximumHeight(300)
+        self.setContextMenuPolicy(Qt.CustomContextMenu)
+        self.customContextMenuRequested.connect(create_menu)
+        self.setUniformRowHeights(True)
+        # remove left margin
+        self.setRootIsDecorated(False)
+        self.setColumnCount(2)
+        self.header().setStretchLastSection(False)
+        sm = QHeaderView.ResizeToContents
+        self.header().setSectionResizeMode(0, sm)
+        self.header().setSectionResizeMode(1, sm)
+
 if __name__ == "__main__":
     app = QApplication([])
     t = WaitingDialog(None, 'testing ...', lambda: [time.sleep(1)], lambda x: QMessageBox.information(None, 'done', "done"))
